import logging
import os
import sys
import tempfile
from glob import glob

import pytest

from bookshelf import BookShelf
from bookshelf.errors import UnknownBook
from bookshelf.notebook import (
    get_available_versions,
    get_notebook_directory,
    run_notebook,
)

logger = logging.getLogger("test-notebooks")


if sys.platform.startswith("win"):
    # https://gitlab.com/climate-resource/bookshelf/bookshelf/-/issues/23
    pytest.skip(
        "skipping notebook tests on windows, see issue !23", allow_module_level=True
    )


def find_notebooks():
    NOTEBOOK_DIRECTORY = get_notebook_directory()

    logger.info(f"Looking for notebooks in {NOTEBOOK_DIRECTORY}")
    notebooks = glob(os.path.join(NOTEBOOK_DIRECTORY, "**", "*.py"), recursive=True)

    notebook_info = []

    for nb in notebooks:
        versions = get_available_versions(
            nb.replace(".py", ".yaml"), include_private=False
        )
        notebook_name = os.path.basename(nb)[:-3]
        notebook_info.extend((nb, notebook_name, v) for v in versions)

    logger.info(f"Found {len(notebook_info)} notebooks")
    for _, name, version in notebook_info:
        logger.info(f"Found {name}@{version}")
    return notebook_info


notebooks = find_notebooks()


@pytest.fixture()
def output_directory():
    out_dir = os.environ.get("BOOKSHELF_CACHE_LOCATION", tempfile.mkdtemp())

    yield out_dir


@pytest.mark.parametrize("notebook_path,notebook_name,notebook_version", notebooks)
def test_notebook(notebook_path, notebook_name, notebook_version, output_directory):
    # Check that:
    # * notebooks run as expected
    # * that hash matches an existing notebook

    notebook_dir = os.path.dirname(notebook_path)

    run_notebook_and_check_results(
        notebook_name,
        version=notebook_version,
        notebook_dir=notebook_dir,
        output_directory=os.path.join(
            output_directory, "sample", notebook_name, notebook_version
        ),
    )


def run_notebook_and_check_results(notebook, version, notebook_dir, output_directory):
    shelf = BookShelf()

    try:
        target_book = run_notebook(
            notebook,
            nb_directory=notebook_dir,
            output_directory=output_directory,
            version=version,
        )
    except UnknownBook:
        logger.info("Book has not been pushed yet")
        return

    if shelf.is_available(name=target_book.name, version=target_book.version):
        existing_book = shelf.load(
            name=target_book.name, version=target_book.version, force=True
        )
        logger.info(f"Remote book exists. Expecting hash: {existing_book.hash()}")

        if existing_book.edition != target_book.edition:
            raise ValueError(
                f"Edition of calculated book doesn't match the remote bookshelf "
                f"({target_book.edition} != {existing_book.edition})"
            )

        if existing_book.hash() != target_book.hash():
            raise ValueError(
                f"Hash of calculated book doesn't match the remote bookshelf "
                f"({target_book.hash()} != {existing_book.hash()})"
            )
    else:
        logger.info("Matching book version does not exist on remote bookshelf")



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # CEDS
#
#
# We want to import by_region and by_region_sector

# %%
import fnmatch
import logging
import tempfile
import zipfile

import pandas as pd
import pooch
import pycountry
import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# #  Initialise

# %%
logging.basicConfig(level="INFO")


# %% tags=["parameters"]
local_bookshelf = tempfile.mkdtemp()
version = "v2016_07_26"

# %%
metadata = load_nb_metadata("ceds", version=version)
metadata.dict()

# %%
book = LocalBook.create_from_metadata(metadata, local_bookshelf=local_bookshelf)

# %% [markdown]
# #  Fetch data using pooch

# %%
ceds_fname = metadata.download_file()
ceds_data = zipfile.ZipFile(ceds_fname)
ceds_archive_fnames = [
    "./" + info.filename
    for info in ceds_data.filelist
    if "country_fuel" not in info.filename
]
ceds_archive_fnames

# %%
if version == "v2016_07_26":
    ceds_species = ["BC", "CH4", "CO2", "CO", "NH3", "NMVOC", "NOx", "OC", "SO2"]
else:
    ceds_species = ["BC", "CH4", "CO2", "CO", "N2O", "NH3", "NMVOC", "NOx", "OC", "SO2"]


# %% [markdown]
# # Process

# %%


def _fix_units(data: scmdata.ScmRun) -> scmdata.ScmRun:
    data["variable"] = data["variable"].str.replace("SO2", "Sulfur")
    data["variable"] = data["variable"].str.replace("NMVOC", "VOC")
    data["unit"] = data["unit"].str.replace("NMVOC", "VOC")

    return (
        data.set_meta("unit", "kt NOx/yr", variable="Emissions|NOx", log_if_empty=False)
        .set_meta("unit", "kt OC/yr", variable="Emissions|OC", log_if_empty=False)
        .set_meta("unit", "kt BC/yr", variable="Emissions|BC", log_if_empty=False)
    )


def read_CEDS_format(fname: str) -> scmdata.ScmRun:
    fname_match = fnmatch.filter(ceds_archive_fnames, "*/" + fname)

    if len(fname_match) != 1:
        raise ValueError(f"Could not figure out match: {fname} -> {fname_match}")
    fname = fname_match[0].lstrip("./")

    df = pd.read_csv(ceds_data.open(fname)).rename(
        {"em": "variable", "country": "region", "units": "unit"}, axis=1
    )
    df["scenario"] = "Historical"
    df["model"] = f"CEDS ({metadata.version})"
    df["unit"] = df["unit"] + "/yr"
    df["unit"] = df["unit"].str.replace("kt", "kt ", regex=False)
    df["variable"] = "Emissions|" + df["variable"]
    df["region"] = df["region"].str.upper()

    columns = []
    for c in df.columns:
        if c.startswith("X"):
            columns.append(int(c[1:]))
        else:
            columns.append(c)
    df.columns = columns

    data = scmdata.ScmRun(df)
    data["region"] = data["region"].str.replace("GLOBAL", "World")
    return _fix_units(data)


# %%
ceds_by_country = []

for species in ceds_species:
    ceds_by_country.append(
        read_CEDS_format(f"{species}_CEDS_emissions_by_country_*.csv")
    )
ceds_by_country = scmdata.run_append(ceds_by_country)

# %%
# Check units
ceds_by_country.meta[["variable", "unit"]].drop_duplicates()

# %%
# Check country codes
for c in ceds_by_country.get_unique_meta("region"):
    if pycountry.countries.get(alpha_3=c) is None:
        print(c)

# %%
ceds_by_country

# %%
book.add_timeseries("by_country", ceds_by_country)


# %% [markdown]
# # By country by sector

# %%
ceds_by_sector = []

for species in ceds_species:
    ceds_by_sector.append(
        read_CEDS_format(f"{species}_CEDS_emissions_by_sector_country_*.csv")
    )
ceds_by_sector = scmdata.run_append(ceds_by_sector)
ceds_by_sector.get_unique_meta("region")

# %%
ceds_by_sector.get_unique_meta("sector")

# %%
ceds_by_sector

# %% [markdown]
# # By grid sectors

# %%
ceds_mapping_fname = pooch.retrieve(
    "https://github.com/JGCRI/CEDS/raw/April-21-2021-release/input/gridding/gridding_mappings/CEDS_sector_to_gridding_sector_mapping.csv",
    known_hash="95f66a04095b3d9f6464d7a4713093ff2967c8a5f386d7b6addad30c40ff12d3",
)
ceds_sector_mapping = pd.read_csv(ceds_mapping_fname)
ceds_sector_mapping


# %%
def process_aggregate_sector(sector_column: str, sector: str):
    target_sector_info = ceds_sector_mapping[
        ceds_sector_mapping[sector_column] == sector
    ]
    ceds_sector_aggregate = ceds_by_sector.filter(
        sector=target_sector_info.CEDS_working_sector.to_list(), log_if_empty=False
    ).process_over(("sector"), "sum")
    ceds_sector_aggregate["sector"] = sector
    ceds_sector_aggregate["sector_short"] = target_sector_info[
        sector_column + "_short"
    ].unique()[0]
    return scmdata.ScmRun(ceds_sector_aggregate)


def extract_sectors(sector_column):
    target_sectors = ceds_sector_mapping[sector_column].unique()

    agg_sectors = []
    for sector in target_sectors:
        if isinstance(sector, str):
            agg_sectors.append(process_aggregate_sector(sector_column, sector))
    return scmdata.run_append(agg_sectors)


# %%
ceds_agg_sectors_intermediate = extract_sectors("CEDS_int_gridding_sector")
ceds_agg_sectors_intermediate.get_unique_meta("sector_short")

# %% [markdown]
# ### Final sectors
#
# aka CEDS9

# %%
ceds_agg_sectors_final = extract_sectors("CEDS_final_gridding_sector")
ceds_agg_sectors_final.get_unique_meta("sector_short")

# %%
ceds_agg_sectors_final

# %%
book.add_timeseries("by_sector_ipcc", ceds_by_sector)
book.add_timeseries("by_sector_intermediate", ceds_agg_sectors_intermediate)
book.add_timeseries("by_sector_final", ceds_agg_sectors_final)

# %% [markdown]
# # Checks

# %%
book.metadata()

# %%



import os

import datapackage
import pooch.hashes
import pytest
import scmdata.testing

from bookshelf.book import LocalBook
from bookshelf.constants import DATA_FORMAT_VERSION, TEST_DATA_DIR
from bookshelf.shelf import BookShelf


def test_book_info():
    book = LocalBook("missing", "v1.1.0")
    assert book.name == "missing"
    assert book.version == "v1.1.0"


def test_book_without_metadata():
    book = LocalBook("missing", "v1.0.0")
    with pytest.raises(FileNotFoundError):
        book.metadata()


def test_create_local(local_bookshelf):
    expected_fname = local_bookshelf / "test" / "v1.0.0_e001" / "datapackage.json"

    assert not expected_fname.exists()
    LocalBook.create_new("test", "v1.0.0", local_bookshelf=local_bookshelf)
    assert expected_fname.exists()


def test_add_timeseries(local_bookshelf, example_data):
    book = LocalBook.create_new("test", "v1.1.0", local_bookshelf=local_bookshelf)
    book.add_timeseries("test", example_data)
    assert len(book.as_datapackage().resources) == 1

    expected_fname = local_bookshelf / "test" / "v1.1.0_e001" / "test.csv"
    assert expected_fname.exists()

    res = book.as_datapackage().resources[0]
    assert res.name == "test"
    assert res.descriptor["format"] == "CSV"
    assert res.descriptor["filename"] == "test.csv"
    assert res.descriptor["hash"] == pooch.hashes.file_hash(expected_fname)


def test_timeseries(example_data):
    book = LocalBook.create_new("test", "v1.1.0")
    book.add_timeseries("test", example_data)

    scmdata.testing.assert_scmdf_almost_equal(example_data, book.timeseries("test"))

    with pytest.raises(ValueError, match="Unknown timeseries 'other'"):
        book.timeseries("other")


def test_timeseries_remote(example_data, remote_bookshelf):
    book = BookShelf().load("test", "v1.0.0")

    scmdata.testing.assert_scmdf_almost_equal(
        example_data, book.timeseries("leakage_rates_low")
    )

    with pytest.raises(ValueError, match="Unknown timeseries 'other'"):
        book.timeseries("other")


def test_metadata():
    book = LocalBook(
        "example",
        "v1.0.0",
        local_bookshelf=os.path.join(TEST_DATA_DIR, DATA_FORMAT_VERSION),
    )

    package = book.as_datapackage()

    assert isinstance(package, datapackage.Package)

    meta_dict = package.descriptor
    assert meta_dict["name"] == book.name
    assert meta_dict["version"] == book.version

    assert meta_dict == book.metadata()


def test_metadata_missing():
    book = LocalBook("example", "v1.0.0")
    with pytest.raises(FileNotFoundError):
        book.metadata()


def test_files(local_bookshelf):
    book = LocalBook("example", "v1.0.0")

    assert len(book.files()) == 0

    book = LocalBook(
        "example",
        "v1.0.0",
        local_bookshelf=os.path.join(TEST_DATA_DIR, DATA_FORMAT_VERSION),
    )
    assert len(book.files()) == 2

    book = LocalBook.create_new("example", "v1.1.0", local_bookshelf=local_bookshelf)
    book_files = book.files()
    assert len(book_files) == 1
    assert book_files[0] == os.path.join(book.local_fname("datapackage.json"))



"""
Bookshelf utilities
"""
import logging
import os
import pathlib
from typing import Any, Optional, Union

import appdirs
import pooch

from bookshelf.constants import DATA_FORMAT_VERSION, DEFAULT_BOOKSHELF, ENV_PREFIX

logger = logging.getLogger(__file__)


def default_cache_location() -> str:
    r"""
    Determine the default cache location


    By default, local Books are stored in the default cache location unless overridden for
    a given :class:`bookshelf.BookShelf`. The default cache location is determined using
    the ``BOOKSHELF_CACHE_LOCATION`` or if that environment variable is not present, it
    falls back to an operating specific location. This location is determined using
    `appdirs <https://github.com/ActiveState/appdirs>`__ and may look like the following:

    * Mac: ``~/Library/Caches/bookshelf``
    * Unix: ``~/.cache/bookshelf`` or the value of the ``XDG_CACHE_HOME``
      environment variable, if defined.
    * Windows: ``C:\Users\<user>\AppData\Local\bookshelf\Cache``

    Returns
    -------
    str
        The default cache location
    """
    return os.environ.get(
        "BOOKSHELF_CACHE_LOCATION",
        appdirs.user_cache_dir("bookshelf", appauthor=False),
    )


def create_local_cache(path: Union[str, pathlib.Path, None] = None) -> pathlib.Path:
    """
    Prepare a cache directory

    Creates a writeable directory in a platform-specific way (see pooch.utils.os_cache)

    Parameters
    ----------
    path: str or path
        If provided, override the default cache location

    Returns
    -------
    pathlib.Path

    Location of a writable local cache
    """
    if path is None:
        path = default_cache_location()

    path = pooch.utils.cache_location(pathlib.Path(path) / DATA_FORMAT_VERSION)

    pooch.utils.make_local_storage(path)

    return pathlib.Path(path)  # type: ignore


def download(
    url: str,
    local_fname: pathlib.Path,
    known_hash: Optional[str] = None,
    progressbar: bool = False,
    retry_count: int = 0,
) -> None:
    """
    Download a remote file using pooch's downloader

    Parameters
    ----------
    url: str
        URL to download
    local_fname
        Path where the result will be stored
    known_hash
    progressbar: bool
        If true, show a progress bar showing the download process
    retry_count: int
        The number of retries to attempt
    """
    downloader = pooch.core.choose_downloader(url, progressbar=progressbar)
    pooch.core.stream_download(
        url,
        fname=local_fname,
        known_hash=known_hash,
        downloader=downloader,
        retry_if_failed=retry_count,
    )


def build_url(bookshelf: str, *paths: str) -> str:
    """
    Build a URL

    Parameters
    ----------
    bookshelf: str
        The remote bookshelf
    paths : list of str
        A collection of paths that form the path of the URL
    Returns
    -------
    str
        The merged URL

    """
    return "/".join([bookshelf, *paths])


def fetch_file(
    url: str,
    local_fname: pathlib.Path,
    known_hash: Optional[str] = None,
    force: Optional[bool] = False,
) -> None:
    """
    Fetch a remote file and store it locally

    Parameters
    ----------
    url : str
        URL of data file
    local_fname : pathlib.Path
        The location of where to store the downloaded file
    known_hash : str
        Expected sha256 has of the output file.

        If the hash of the downloaded file doesn't match the expected hash a ValueError
        is raised.

        If no hash is provided, no checks are performed
    force : bool
        If True, always download the file

    Raises
    ------
    ValueError
        Failing hash check for the output file
    FileNotFoundError
        Downloaded file was not in the expected location

    """
    if not force and local_fname.exists():
        if pooch.hashes.hash_matches(local_fname, known_hash):
            return
        raise ValueError(
            f"Hash for existing file {local_fname} does not match the expected "
            f"value {known_hash}"
        )

    if force or not local_fname.exists():
        download(url, local_fname=local_fname, known_hash=known_hash)
        logger.info(f"{local_fname} downloaded from {url}")

    if not local_fname.exists():
        raise FileNotFoundError(
            f"Could not find file {local_fname}"
        )  # pragma: no cover


def get_env_var(
    name: str,
    add_prefix: bool = True,
    raise_on_missing: bool = True,
    default: Any = None,
) -> str:
    """
    Get an environment variable value

    If the variable isn't set raise an exception if ``raise_on_missing`` is ``True``

    Parameters
    ----------
    name : str
        Environment variable name to check
    add_prefix : bool
        If ``True``, prefix the environment variable name with  ``bookshelf.constants.ENV_PREFIX``
    raise_on_missing : bool
        If ``True``, a ValueError is raised
    default : Any
        Value to return if the environment variable is missing and ``raise_on_missing`` is ``True``
    Returns
    -------
    str
        Value of environment variable

    Raises
    ------
    ValueError
        Environment variable not set
    """
    if add_prefix:
        name = ENV_PREFIX + name
    name = name.upper()
    if raise_on_missing and name not in os.environ:
        raise ValueError(f"Environment variable {name} not set. Check configuration")
    return os.environ.get(name, default)


def get_remote_bookshelf(bookshelf: Optional[str]) -> str:
    """
    Get the remote bookshelf URL

    If no bookshelf is provided, use the ``BOOKSHELF_REMOTE`` environment variable, or,
    the ``bookshop.constants.DEFAULT_BOOKSHELF`` parameter if the environment variable
    is not present.

    Parameters
    ----------
    bookshelf : str
        URL for the bookshelf

        If not provided the URL is determined as above

    Returns
    -------
    str
        URL for the remote bookshelf
    """
    if bookshelf is None:
        return os.environ.get(ENV_PREFIX + "REMOTE", DEFAULT_BOOKSHELF)
    return bookshelf



import os.path
import re
import tempfile

import pytest

import bookshelf.notebook
from bookshelf.errors import UnknownVersion
from bookshelf.notebook import get_notebook_directory, load_nb_metadata, run_notebook


def test_load_nb_metadata():
    res = load_nb_metadata(
        "simple",
        nb_directory=os.path.join(get_notebook_directory(), "examples", "simple"),
    )

    assert res.name == "simple"


def test_load_nb_metadata_versioned():
    # Use the default version suffix
    res = load_nb_metadata(
        "multiple_versions",
        version="v5.1.0",
        nb_directory=os.path.join(
            get_notebook_directory(), "examples", "multiple_versions"
        ),
    )
    assert res.name == "multiple_versions"
    assert res.version == "v5.1.0"

    # Be explicit
    res = load_nb_metadata(
        "multiple_versions",
        nb_directory=os.path.join(
            get_notebook_directory(), "examples", "multiple_versions"
        ),
        version="v4.0.0",
    )
    assert res.name == "multiple_versions"
    assert res.version == "v4.0.0"

    with pytest.raises(UnknownVersion):
        load_nb_metadata(
            "examples/multiple_versions",
            version="v1.0.0",
        )


def test_load_nb_metadata_paths():
    simple_nb_dir = os.path.join(get_notebook_directory(), "examples", "simple")

    load_nb_metadata("simple", nb_directory=simple_nb_dir)
    load_nb_metadata("examples/simple")
    load_nb_metadata("examples/simple/simple.yaml")
    exp = load_nb_metadata(
        "simple",
        nb_directory=os.path.join(get_notebook_directory(), "examples"),
    )
    assert exp
    assert load_nb_metadata("simple.yaml", nb_directory=simple_nb_dir) == exp
    nb_dir = os.path.join(simple_nb_dir, "simple.yaml")
    assert load_nb_metadata(os.path.abspath(nb_dir)) == exp


@pytest.mark.xfail(reason="missing file")
def test_load_nb_private():
    exp = load_nb_metadata("primap-hist", "v2.4_alt")

    assert exp.private

    assert not load_nb_metadata("primap-hist", "v2.4").private


@pytest.mark.parametrize(
    "name,version",
    [
        ("simple", "v0.1.0"),
        ("multiple_versions", "v4.0.0"),
        ("multiple_versions", "v5.1.0"),
    ],
)
def test_run_notebook(remote_bookshelf, name, version):
    nb_name = f"examples/{name}"
    remote_bookshelf.register(name, version, 1)

    with tempfile.TemporaryDirectory() as td:
        book = run_notebook(
            nb_name,
            output_directory=td,
            version=version,
        )

        with pytest.raises(
            ValueError, match=re.escape(f"{os.path.join(td, version)} is not empty")
        ):
            run_notebook(
                nb_name,
                output_directory=td,
                version=version,
            )
        run_notebook(
            f"examples/{name}",
            output_directory=td,
            force=True,
            version=version,
        )

        assert len(book.files()) == 2
        assert os.listdir(td) == [version]
        assert sorted(os.listdir(os.path.join(td, version))) == sorted(
            [
                name,  # output bookshelf
                f"{name}.ipynb",
                f"{name}.yaml",
                f"{name}.py",
            ]
        )


def test_run_missing_version():
    with pytest.raises(UnknownVersion, match="Could not find simple@v200000"):
        run_notebook(
            "examples/simple",
            version="v200000",
        )


@pytest.mark.parametrize("package", ["jupytext", "papermill"])
def test_missing_deps(package):
    try:
        setattr(bookshelf.notebook, f"has_{package}", False)

        match = re.escape(
            f"{package} is not installed. Run 'pip install bookshelf[notebooks]'"
        )
        with pytest.raises(ImportError, match=match):
            run_notebook("test")
    finally:
        bookshelf.notebook.has_papermill = True
        bookshelf.notebook.has_jupytext = True



"""
Schema
"""
from typing import Any, Dict, List, Optional

import pooch
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

from bookshelf.utils import get_env_var

Version = str
Edition = int


class BookVersion(BaseModel):
    """
    Version information for a book
    """

    version: Version
    edition: Edition
    url: str
    hash: str
    private: Optional[bool] = False


class VolumeMeta(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)
    """

    name: str
    license: str  # A change in license will require a new volume
    versions: List[BookVersion]

    def get_latest_version(self) -> Version:
        """
        Get the latest version for a volume

        Returns
        -------
        Version string
        """
        ordered_versions = sorted([v.version for v in self.versions if not v.private])
        if not ordered_versions:
            raise ValueError("No published volumes")

        return ordered_versions[-1]

    def get_version(self, version: Version) -> List[BookVersion]:
        """
        Get a set of books for a given version

        Returns
        -------
        List of matching books sorted by edition
        """
        matching_versions = []

        for version_meta in self.versions:
            if version_meta.version == version:
                matching_versions.append(version_meta)

        matching_versions = sorted(matching_versions, key=lambda v: v.edition)
        return matching_versions


class FileDownloadInfo(BaseModel):
    """
    A File to be downloaded as part of a dataset
    """

    url: str
    hash: str


class DatasetMetadata(BaseModel):
    """
    Metadata about a dataset

    A dataset may consist of multiple files (:class:`FileDownloadInfo`)
    """

    url: Optional[str]
    doi: Optional[str]
    files: List[FileDownloadInfo]
    author: str


class VersionMetadata(BaseModel):
    """
    Metadata about a single version of a book
    """

    version: Version
    dataset: DatasetMetadata
    private: Optional[bool] = Field(default=False)


class NotebookMetadata(BaseModel):
    """
    Schema for a running a notebook

    This represents the metadata for a selected version of a volume
    """

    name: str
    version: Version
    edition: Edition
    description: Optional[str]
    license: str
    source_file: str
    private: bool
    metadata: Dict[str, Any]  # TODO: type this
    dataset: DatasetMetadata

    def long_name(self) -> str:
        """
        Long name of the book

        Includes name and long version

        Returns
        -------
        str
            Long identifier for a book
        """
        return f"{self.name}@{self.long_version()}"

    def long_version(self) -> str:
        """
        Long version identifier

        Of the form "{version}_e{edition}" e.g. v1.0.1_e002.

        Returns
        -------
        str
            Version identification string
        """
        return f"{self.version}_e{self.edition:03}"

    def download_file(self, idx: int = 0) -> str:
        """
        Download a dataset file

        Uses ``pooch`` to manage the downloading, verification and caching of data file.
        The first call will trigger a download and subsequent calls may use the cached
        file if the previous download succeeded.

        Parameters
        ----------
        idx
            Index of the file to download (0-based). Defaults to the first file if
            no value is provided

        Returns
        -------
        str
            Filename of the locally downloaded file
        """
        cache_location = get_env_var(
            "DOWNLOAD_CACHE_LOCATION", raise_on_missing=False, default=None
        )
        file_info = self.dataset.files[idx]

        file_hash: Optional[str] = file_info.hash
        if not file_hash:
            # replace an empty string with None
            file_hash = None

        if file_info.url.startswith("file://"):
            return file_info.url[7:]

        res: str = pooch.retrieve(
            file_info.url,
            known_hash=file_hash,
            path=cache_location,
        )
        return res


class ConfigSchema(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)

    A volume can hold multiple versions of the same data
    """

    name: str
    edition: Edition
    description: Optional[str]
    license: str
    source_file: str
    metadata: Dict[str, Any]  # TODO: type this
    versions: List[VersionMetadata]



import io
import json
import logging
import os
import pathlib
import platform
import re

import appdirs
import boto3
import boto3.exceptions
import moto
import pytest

from bookshelf.constants import DATA_FORMAT_VERSION
from bookshelf.errors import UnknownBook, UnknownVersion, UploadError
from bookshelf.notebook import NotebookMetadata
from bookshelf.shelf import BookShelf, LocalBook, _update_volume_meta


@pytest.fixture()
def shelf(local_bookshelf):
    shelf = BookShelf(path=local_bookshelf)
    return shelf


def get_meta(conn, bucket, pkg):
    volume_meta_contents = io.BytesIO()
    conn.Object(bucket, f"/this/prefix/{pkg}/volume.json").download_fileobj(
        volume_meta_contents
    )
    volume_meta_contents.seek(0)
    return json.load(volume_meta_contents)


def setup_upload_bucket(remote_bookshelf, monkeypatch):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/new-package/volume.json", status_code=404
    )

    bucket = "test-bucket"
    prefix = "/this/prefix"
    monkeypatch.setenv("BOOKSHELF_BUCKET", bucket)
    monkeypatch.setenv("BOOKSHELF_BUCKET_PREFIX", prefix)

    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    return conn


def test_local_cache(monkeypatch):
    # Override the local_bookshelf fixture
    monkeypatch.delenv("BOOKSHELF_CACHE_LOCATION")

    shelf = BookShelf()

    if platform.system() == "Windows":
        exp = (
            pathlib.Path(appdirs.user_cache_dir())
            / "bookshelf"
            / "Cache"
            / DATA_FORMAT_VERSION
        )
    else:
        exp = pathlib.Path(appdirs.user_cache_dir()) / "bookshelf" / DATA_FORMAT_VERSION
    assert shelf.path == exp


def test_load_missing(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/missing/volume.json", status_code=404
    )

    with pytest.raises(UnknownBook, match=re.escape("No metadata for 'missing'")):
        shelf.load("missing")


def test_load_missing_version(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/test/v1.1.1/datapackage.json", status_code=404
    )
    with pytest.raises(UnknownVersion, match="Could not find test@v1.1.1"):
        shelf.load("test", "v1.1.1")

    with pytest.raises(UnknownVersion, match="Could not find test@v1.1.1"):
        shelf.load("test", "v1.1.1", force=True)


def test_load(remote_bookshelf, shelf):
    book = shelf.load("test")
    assert book.version == "v1.1.0"

    book = shelf.load("test", "v1.1.0")
    assert book.version == "v1.1.0"

    book = shelf.load("test", "v1.0.0")
    assert book.version == "v1.0.0"


def test_load_with_cached(remote_bookshelf, local_bookshelf):
    shelf = BookShelf(path=local_bookshelf)
    shelf.load("test")

    remote_bookshelf.mocker.reset()
    remote_bookshelf.register("test", "v1.1.0", 2)
    remote_bookshelf.register("test", "v1.2.1", 1)

    shelf = BookShelf(path=local_bookshelf)
    shelf.load("test", "v1.1.0", 1)
    assert remote_bookshelf.mocker.call_count == 0

    # Not providing an edition causes a fetch of the volume and datapackage
    book = shelf.load("test", "v1.1.0")
    assert book.edition == 2
    assert remote_bookshelf.mocker.call_count == 2

    # A cache miss causes a fetch of the data package
    shelf.load("test", "v1.2.1", 1)
    assert remote_bookshelf.mocker.call_count == 3

    # Not providing a version/edition always causes a fetch
    shelf.load("test")
    assert remote_bookshelf.mocker.call_count == 4


@moto.mock_s3
def test_publish(shelf, remote_bookshelf, monkeypatch, caplog, example_data):
    conn = setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("new-package", "v1.1.1")
    book.add_timeseries("example", example_data)

    with caplog.at_level(logging.INFO):
        shelf.publish(book)

    # Check that files uploaded
    bucket = os.environ["BOOKSHELF_BUCKET"]
    conn.Object(bucket, "/this/prefix/new-package/v1.1.1_e001/datapackage.json").load()
    assert (
        conn.Object(bucket, "/this/prefix/new-package/v1.1.1_e001/datapackage.json")
        .Acl()
        .grants[1]["Permission"]
        == "READ"
    )
    conn.Object(bucket, "/this/prefix/new-package/v1.1.1_e001/example.csv").load()

    volume_meta = get_meta(conn, bucket, "new-package")
    assert len(volume_meta["versions"]) == 1
    assert volume_meta["versions"][-1]["version"] == "v1.1.1"

    assert "Book new-package@v1.1.1 ed.1 uploaded successfully" in caplog.text


@moto.mock_s3
def test_publish_new_version(
    shelf, remote_bookshelf, monkeypatch, caplog, example_data
):
    conn = setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("test", "v1.1.1", edition=25)

    with caplog.at_level(logging.INFO):
        shelf.publish(book)

    # Check that files uploaded
    bucket = os.environ["BOOKSHELF_BUCKET"]
    conn.Object(bucket, "/this/prefix/test/v1.1.1_e025/datapackage.json").load()

    volume_meta = get_meta(conn, bucket, "test")
    assert len(volume_meta["versions"]) == 3  # 2 existing in volume.json + one extra
    assert volume_meta["versions"][-1]["version"] == "v1.1.1"

    assert "Book test@v1.1.1 ed.25 uploaded successfully" in caplog.text


@pytest.mark.parametrize("edition", [2, 900])
@moto.mock_s3
def test_publish_new_edition(
    shelf, remote_bookshelf, monkeypatch, caplog, example_data, edition
):
    conn = setup_upload_bucket(remote_bookshelf, monkeypatch)
    book = LocalBook.create_new("test", "v1.1.0", edition=edition)

    with caplog.at_level(logging.INFO):
        shelf.publish(book)

    # Check that files uploaded
    bucket = os.environ["BOOKSHELF_BUCKET"]
    conn.Object(
        bucket, f"/this/prefix/test/v1.1.0_e{edition:03}/datapackage.json"
    ).load()

    volume_meta = get_meta(conn, bucket, "test")
    assert len(volume_meta["versions"]) == 3  # 2 existing in volume.json + one extra
    assert volume_meta["versions"][-1]["version"] == "v1.1.0"
    assert volume_meta["versions"][-1]["edition"] == edition

    assert "Uploading a new edition of an existing book" in caplog.text
    assert f"Book test@v1.1.0 ed.{edition} uploaded successfully" in caplog.text


@moto.mock_s3
def test_publish_wrong_permissions(shelf, remote_bookshelf, monkeypatch, caplog):
    setup_upload_bucket(remote_bookshelf, monkeypatch)

    # Modify bucket used (this bucket hasn't been created)
    bucket = "test-bucket-wrong"
    monkeypatch.setenv("BOOKSHELF_BUCKET", bucket)

    caplog.set_level(logging.ERROR)

    book = LocalBook.create_new("new-package", "v1.0.0")
    with pytest.raises(
        UploadError, match=re.escape(f"Failed to upload {book.files()[0]} to s3")
    ):
        shelf.publish(book)

    assert "NoSuchBucket" in caplog.text
    assert bucket in caplog.text


def test_publish_existing(shelf, remote_bookshelf):
    book = LocalBook.create_new("test", "v1.0.0", 1)

    with pytest.raises(
        UploadError,
        match=re.escape(
            "Edition value has not been increased (remote: v1.0.0_e001, local: v1.0.0_e001)"
        ),
    ):
        shelf.publish(book, force=False)


@moto.mock_s3
def test_publish_existing_forced(shelf, remote_bookshelf, monkeypatch, caplog):
    setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("new-package", "v1.0.0", 1)

    with caplog.at_level(logging.INFO):
        shelf.publish(book, force=True)

    assert "Book new-package@v1.0.0 ed.1 uploaded successfully" in caplog.text


def test_publish_extra_file(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/new-package/volume.json", status_code=404
    )

    book = LocalBook.create_new("new-package", "v1.0.0")
    open(book.local_fname("extra_file.txt"), "w").close()

    with pytest.raises(
        UploadError, match="Non-resource file extra_file.txt found in book"
    ):
        shelf.publish(book)


def test_is_available(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/other/volume.json", status_code=404
    )

    assert shelf.is_available("test", "v1.0.0")
    assert shelf.is_available("test", "v1.1.0")

    assert not shelf.is_available("test", "v1.1.1")
    assert not shelf.is_available("other", "v1.1.0")


def test_is_available_any_version(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/other/volume.json", status_code=404
    )

    assert shelf.is_available("test")
    assert not shelf.is_available("other")


def test_is_cached(shelf):
    book = LocalBook.create_new("test", "v1.0.0", edition=1, local_bookshelf=shelf.path)

    assert shelf.is_cached(book.name, book.version, edition=1)
    assert not shelf.is_cached(book.name, "v1.0.1", 1)
    assert not shelf.is_cached("other", "v1.0.1", 1)


def test_list_versions(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get(
        f"/{DATA_FORMAT_VERSION}/other/volume.json", status_code=404
    )

    assert shelf.list_versions("test") == ["v1.0.0", "v1.1.0"]
    with pytest.raises(UnknownBook):
        shelf.list_versions("other")


@pytest.mark.xfail(reason="Not implemented")
def test_list_name(shelf):
    assert shelf.list_books() == ["test"]


def test_private_list(remote_bookshelf, local_bookshelf):
    remote_bookshelf.register("test", "v2_private", 1, private=True)
    shelf = BookShelf(path=local_bookshelf)

    assert shelf.list_versions("test") == ["v1.0.0", "v1.1.0"]
    assert shelf.load("test", "v2_private")

    assert shelf.load("test").version == "v1.1.0"


def test_update_volume_meta(local_bookshelf, remote_bookshelf):
    book = LocalBook.create_from_metadata(
        NotebookMetadata(
            name="test",
            version="v2_private",
            edition=1,
            license="",
            source_file="",
            private=True,
            dataset={"author": "", "files": []},
            metadata={},
        ),
        local_bookshelf=local_bookshelf,
    )

    res = _update_volume_meta(book, os.environ["BOOKSHELF_REMOTE"])
    with open(res) as fh:
        contents = json.load(fh)

    assert contents["versions"][-1]["version"] == "v2_private"
    assert contents["versions"][-1]["private"]



"""
Custom exceptions
"""


class UnknownBook(ValueError):
    """
    An unknown book is requested
    """


class UnknownVersion(ValueError):
    """
    An unknown version is requested
    """

    def __init__(self, name, version):
        self.name = name
        self.version = version
        super().__init__()

    def __str__(self):
        return f"Could not find {self.name}@{self.version}"


class UnknownEdition(UnknownVersion):
    """
    An unknown edition is requested
    """

    def __init__(self, name, version, edition):
        super().__init__(name, version)
        self.edition = edition

    def __str__(self):
        return f"Could not find {self.name}@{self.version} ed.{self.version}"


class UploadError(ValueError):
    """
    Could not upload a book to the remote bookshelf
    """



import os
import shutil

import pytest
import scmdata

from bookshelf.constants import TEST_DATA_DIR
from bookshelf.utils import create_local_cache


@pytest.fixture(scope="function", autouse=True)
def local_bookshelf(tmpdir, monkeypatch):
    monkeypatch.setenv("BOOKSHELF_CACHE_LOCATION", str(tmpdir))
    fname = create_local_cache(tmpdir)
    yield fname

    shutil.rmtree(fname)


@pytest.fixture()
def example_data():
    return scmdata.ScmRun(
        scmdata.ScmRun(
            os.path.join(
                TEST_DATA_DIR,
                "v0.2.0",
                "example",
                "v1.0.0_e001",
                "leakage_rates_low.csv",
            )
        )
        .timeseries()
        .sort_index()
    )



"""
run CLI command
"""
import logging
from typing import Tuple

import click

from bookshelf.notebook import get_available_versions, run_notebook

logger = logging.getLogger(__name__)


@click.command("run", short_help="Run a notebook")
@click.argument("name", required=True)
@click.option(
    "-o",
    "--output",
    help="Directory to store the artifacts from running the notebook",
    required=False,
)
@click.option(
    "--version",
    multiple=True,
    help="List of versions to run",
    required=False,
)
@click.option(
    "--include-private/--no-include-private",
    help="Run private versions. These will likely fail if the data is not available locally",
    default=False,
)
@click.option(
    "-f",
    "--force",
    help="Override the existing output if the output directory isn't empty",
    is_flag=True,
)
def cli(
    name: str, output: str, force: bool, version: Tuple[str], include_private: bool
) -> None:
    """
    Run a notebook

    This runs one of the notebooks used to generate a Book
    """
    if not version:
        all_versions = get_available_versions(name, include_private=include_private)
    else:
        all_versions = version

    for dataset_version in all_versions:
        try:
            run_notebook(
                name, output_directory=output, force=force, version=dataset_version
            )
        except Exception as exc:
            logger.error(f"Failed to run {name}: {exc}")
            raise click.Abort() from exc



import logging
import re
import tempfile

import pytest
from click.testing import CliRunner

from bookshelf.cli import main
from bookshelf.shelf import BookShelf

logging.basicConfig()


def test_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0

    assert "Usage: bookshelf [OPTIONS] COMMAND [ARGS]" in result.output
    assert "Bookshelf for managing reusable datasets" in result.output

    expected_commands = [r"run\s+Run a notebook"]

    for exp in expected_commands:
        assert re.search(exp, result.output)


def test_run(mocker):
    mock_run = mocker.patch("bookshelf.commands.cmd_run.run_notebook", autospec=True)

    with tempfile.TemporaryDirectory() as td:
        runner = CliRunner()
        result = runner.invoke(main, ["run", "examples/simple", "--output", str(td)])
        assert result.exit_code == 0, result.output

        mock_run.assert_called_once_with(
            "examples/simple", output_directory=td, force=False, version="v0.1.0"
        )


@pytest.mark.parametrize(
    "args",
    [
        [],
        [
            "--version",
            "v4.0.0",
            "--version",
            "v5.1.0",
        ],
    ],
)
def test_run_multiple(mocker, args):
    mock_run = mocker.patch("bookshelf.commands.cmd_run.run_notebook", autospec=True)

    with tempfile.TemporaryDirectory() as td:
        runner = CliRunner()
        result = runner.invoke(
            main,
            [
                "run",
                "examples/multiple_versions",
                *args,
                "--output",
                str(td),
            ],
        )
        assert result.exit_code == 0, result.output

        mock_run.assert_has_calls(
            [
                (
                    ("examples/multiple_versions",),
                    dict(
                        output_directory=td,
                        force=False,
                        version="v4.0.0",
                    ),
                ),
                (
                    ("examples/multiple_versions",),
                    dict(
                        output_directory=td,
                        force=False,
                        version="v5.1.0",
                    ),
                ),
            ]
        )


def test_run_failed(mocker, caplog):
    mock_run = mocker.patch("bookshelf.commands.cmd_run.run_notebook", autospec=True)
    mock_run.side_effect = ValueError("Something went wrong")
    with tempfile.TemporaryDirectory() as td:
        runner = CliRunner()
        result = runner.invoke(main, ["run", "examples/simple", "--output", str(td)])
        assert result.exit_code == 1

    assert "Something went wrong" in caplog.text
    assert "Aborted!" in result.output


def test_publish(mocker, caplog):
    mock_run = mocker.patch(
        "bookshelf.commands.cmd_publish.run_notebook", autospec=True
    )
    mock_publish = mocker.patch.object(BookShelf, "publish", autospec=True)

    caplog.set_level("INFO")

    runner = CliRunner()
    result = runner.invoke(main, ["publish", "examples/simple"])
    assert result.exit_code == 0

    assert "Building Book in isolated environment" in caplog.text

    mock_run.assert_called_once()
    mock_publish.assert_called_once()

    assert mock_publish.call_args.args[1] == mock_run.return_value


def test_publish_multiple(mocker, caplog):
    mock_run = mocker.patch(
        "bookshelf.commands.cmd_publish.run_notebook", autospec=True
    )
    mock_publish = mocker.patch.object(BookShelf, "publish", autospec=True)

    runner = CliRunner()
    result = runner.invoke(main, ["publish", "examples/multiple_versions"])

    assert result.exit_code == 0

    assert mock_run.call_count == 2
    assert mock_publish.call_count == 2


def test_publish_failed(mocker, caplog):
    mock_run = mocker.patch("bookshelf.commands.cmd_publish.run_notebook")
    mock_run.side_effect = ValueError("Something went wrong")

    runner = CliRunner()
    result = runner.invoke(main, ["publish", "examples/simple"])
    assert result.exit_code == 1

    assert "Something went wrong" in caplog.text
    assert "Aborted!" in result.output


def test_publish_missing():
    runner = CliRunner()
    result = runner.invoke(main, ["publish", "examples/unknown"])
    assert result.exit_code == 1

    assert result.exc_info[0] == FileNotFoundError



import json
import os

import pytest

from bookshelf.constants import DATA_FORMAT_VERSION, TEST_DATA_DIR


def read_json(fname):
    fname = os.path.join(TEST_DATA_DIR, fname)
    with open(fname) as fh:
        return json.load(fh)


def read_data(fname):
    fname = os.path.join(TEST_DATA_DIR, fname)
    with open(fname) as fh:
        return fh.read()


@pytest.fixture(scope="function", autouse=True)
def remote_bookshelf(requests_mock, monkeypatch):
    prefix = f"https://bookshelf.local/{DATA_FORMAT_VERSION}"
    monkeypatch.setenv("BOOKSHELF_REMOTE", prefix)

    class MockRemoteBookshelf:
        def __init__(self):
            self.mocker = requests_mock
            self.meta = {}

            self.register("test", "v1.0.0", 1)
            self.register("test", "v1.1.0", 1)

        def register(self, name, version, edition, private=False):
            if name not in self.meta:
                self.meta[name] = {
                    "name": name,
                    "license": "MIT",
                    "versions": [],
                }

            url_prefix = f"{prefix}/{name}/{version}_e{edition:03}"
            self.meta[name]["versions"].append(
                {
                    "version": version,
                    "edition": edition,
                    "hash": "",
                    "url": url_prefix,
                    "private": private,
                }
            )
            requests_mock.get(
                f"{prefix}/{name}/volume.json",
                json=self.meta[name],
            )
            requests_mock.get(
                f"{url_prefix}/datapackage.json",
                json=read_json("v0.2.0/example/v1.0.0_e001/datapackage.json"),
            )
            requests_mock.get(
                f"{url_prefix}/leakage_rates_low.csv",
                text=read_data("v0.2.0/example/v1.0.0_e001/leakage_rates_low.csv"),
            )

    bs = MockRemoteBookshelf()

    yield bs



from bookshelf import BookShelf


# Fetches from the remote bookshelf
def test_simple_book(local_bookshelf, caplog):
    with caplog.at_level(level="INFO"):
        shelf = BookShelf()
        rcmip_book = shelf.load("rcmip-emissions", "v5.1.0")

    expected_dir = local_bookshelf / "rcmip-emissions"
    # assert (expected_dir / "volume.json").exists()
    assert (expected_dir / "v5.1.0_e001" / "datapackage.json").exists()

    metadata = rcmip_book.metadata()
    assert metadata["version"] == "v5.1.0"
    assert len(metadata["resources"])

    magicc = rcmip_book.timeseries("magicc")
    assert len(magicc)
    assert (expected_dir / "v5.1.0_e001" / "magicc.csv").exists()



"""
Functions to run/manage notebooks
"""
import logging
import os
import shutil
from typing import Any, Dict, Optional, Tuple

# pylint: disable=invalid-name
try:
    import jupytext

    has_jupytext = True
except ImportError:  # pragma: no cover
    jupytext = None
    has_jupytext = False
# pylint: disable=invalid-name
try:
    import papermill

    has_papermill = True
except ImportError:  # pragma: no cover
    papermill = None
    has_papermill = False

import yaml

from bookshelf import BookShelf, LocalBook
from bookshelf.constants import PROCESSED_DATA_DIR, ROOT_DIR
from bookshelf.errors import UnknownVersion
from bookshelf.schema import ConfigSchema, NotebookMetadata, Version
from bookshelf.utils import get_env_var

logger = logging.getLogger(__name__)


def get_notebook_directory(nb_dir: Optional[str] = None) -> str:
    """
    Get the root location of the notebooks used to generate books

    The order of lookup is (in increasing precedence):

    * default ("/notebooks" in a locally checked out version of the repository)
    * ``BOOKSHELF_NOTEBOOK_DIRECTORY`` environment variable
    * ``nb_dir`` parameter

    Parameters
    ----------
    nb_dir : str
        If provided override the default value

    Returns
    -------
    str
        Location of notebooks

    """
    if nb_dir:
        return nb_dir

    try:
        nb_directory = get_env_var("NOTEBOOK_DIRECTORY")
    except ValueError:
        nb_directory = os.path.join(ROOT_DIR, "notebooks")

    return nb_directory


def _load_nb_config(
    name: str,
    nb_directory: Optional[str] = None,
) -> Tuple[ConfigSchema, Dict[str, Any]]:
    nb_directory = get_notebook_directory(nb_directory)

    metadata_fname = name
    if nb_directory and not os.path.isabs(name):
        metadata_fname = os.path.join(nb_directory, name)

    # If a directory is provided assume that the config is similarly named
    if os.path.isdir(metadata_fname):
        metadata_fname = os.path.join(metadata_fname, name.split("/")[-1] + ".yaml")

    if not metadata_fname.endswith(".yaml") and not metadata_fname.endswith(".yml"):
        metadata_fname = metadata_fname + ".yaml"

    if not os.path.exists(metadata_fname):
        raise FileNotFoundError(f"Could not find {metadata_fname}")

    with open(metadata_fname) as file_handle:
        data = yaml.safe_load(file_handle)

        # always override with the file that was used
        data["source_file"] = metadata_fname
        return ConfigSchema(**data), data


def load_nb_metadata(
    name: str,
    version: Optional[Version] = None,
    nb_directory: Optional[str] = None,
) -> NotebookMetadata:
    """
    Load notebook metadata

    Attempts to search :param:`nb_directory` for a metadata YAML file. This YAML file
    contains information about the dataset that is being processed. See NotebookMetadata
    for a description of the available options.

    The assumed filename format for versioned data is {name}_{version}.yaml where
    name matches the notebook name and the name as specified in the NotebookMetadata

    Parameters
    ----------
    name : str
        Filename to load. Should match the notebook name (not checked)
    version : str
        Version of the metadata to load. If none is provided, the last version will be used
    nb_directory: str
        If a non-absolute path is provided, it is assumed to be relative to nb_directory

    Raises
    ------
    UnknownVersion
        A matching version is not in the configuration

    Returns
    -------
    NotebookMetadata
        Metadata about the notebook including the target package and version
    """
    config, raw_data = _load_nb_config(name, nb_directory)

    if "version" in raw_data and version != raw_data["version"]:
        # Check if a version has already been selected
        raise ValueError("Requested version does not match the metadata")

    if version:
        selected_version = None
        for v in config.versions:
            if v.version == version:
                selected_version = v
    else:
        selected_version = config.versions[-1]
    if selected_version is None:
        raise UnknownVersion(config.name, version)

    return NotebookMetadata(**raw_data, **selected_version.dict())


def run_notebook(
    name: str,
    nb_directory: Optional[str] = None,
    output_directory: Optional[str] = None,
    force: bool = False,
    version: Optional[Version] = None,
) -> LocalBook:
    """
    Run a notebook to generate a new Book

    The jupytext ``.py`` version of the notebook is used.

    The template file and configuration is copied to the output directory. The
    template ``.py`` file is then used to create a notebook which is run using
    ``papermill``. The ``local_bookshelf`` parameter is also set to the output
    directory.

    Parameters
    ----------
    name : str
        Name of the notebook
    nb_directory : str
        Directory containing the notebooks.

        This defaults to the ``notebooks/`` directory in this project
    output_directory : str
        Where the output directory will be created.

        This defaults to ``data/processing/{name}/{version}``
    force : bool
        If True, override the existing data in the output directory
    version : str
        Version to extract

    Returns
    -------
    LocalBook
        The generated book
    """
    if not has_papermill:
        raise ImportError(
            "papermill is not installed. Run 'pip install bookshelf[notebooks]'"
        )
    if not has_jupytext:
        raise ImportError(
            "jupytext is not installed. Run 'pip install bookshelf[notebooks]'"
        )

    short_name = name.split("/")[-1]

    # Verify metadata
    metadata = load_nb_metadata(name, version=version, nb_directory=nb_directory)
    nb_fname = metadata.source_file.replace(".yaml", ".py")

    if not os.path.exists(nb_fname):
        raise FileNotFoundError(f"Could not find notebook: {nb_fname}")

    logger.info(f"Loaded metadata from {metadata.source_file}")
    if metadata.name != short_name:  # pragma: no cover
        raise ValueError(
            f"name in metadata does not match the name of the notebook "
            f"({metadata.name} != {name}"
        )
    logger.info(f"Processing {metadata.long_name()}")

    if output_directory is None:
        output_directory = os.path.join(PROCESSED_DATA_DIR, short_name)

    output_directory = os.path.join(output_directory, metadata.version)
    if os.path.exists(output_directory) and os.listdir(output_directory):
        logger.warning(f"{output_directory} is not empty")
        if not force:
            raise ValueError(f"{output_directory} is not empty")
    os.makedirs(output_directory, exist_ok=True)

    # Copy required files
    logger.info(f"Copying {nb_fname} to {output_directory}")
    shutil.copyfile(nb_fname, os.path.join(output_directory, f"{short_name}.py"))
    logger.info(f"Copying metadata to {output_directory}")
    with open(os.path.join(output_directory, f"{short_name}.yaml"), "w") as fh:
        yaml.safe_dump(metadata.dict(), fh)

    # Template and run notebook
    output_nb_fname = os.path.join(output_directory, f"{short_name}.ipynb")
    logger.info(f"Creating notebook {output_nb_fname} from {nb_fname}")
    notebook_jupytext = jupytext.read(nb_fname)
    jupytext.write(
        notebook_jupytext,
        output_nb_fname,
        fmt="ipynb",
    )
    papermill.execute_notebook(
        output_nb_fname,
        output_nb_fname,
        parameters={"local_bookshelf": output_directory, "version": version},
    )
    # Attempt to load the book from the output directory
    shelf = BookShelf(path=output_directory)
    book = shelf.load(short_name, metadata.version, edition=metadata.edition)

    logger.info(f"Notebook run successfully with hash: {book.hash()}")
    return book


def get_available_versions(name: str, include_private: bool = False) -> list[str]:
    """
    Get a list of available versions of a book

    Parameters
    ----------
    name
        Package name
    include_private
        If True, also include private versions

    Returns
    -------
        List of versions
    """
    config, _ = _load_nb_config(name)

    versions = config.versions
    if not include_private:
        versions = [v for v in versions if not v.private]

    return [v.version for v in versions]



"""
A Book represents a single versioned dataset. A dataset can contain multiple resources
each of which are loaded independently.
"""
import glob
import json
import os.path
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import datapackage
import pooch
import scmdata

from bookshelf.schema import Edition, NotebookMetadata, Version
from bookshelf.utils import (
    build_url,
    create_local_cache,
    fetch_file,
    get_remote_bookshelf,
)

DATAPACKAGE_FILENAME = "datapackage.json"


class _Book:
    def __init__(
        self,
        name: str,
        version: str,
        edition: int,
        bookshelf: Optional[str] = None,
    ):
        self.name = name
        self.version = version
        self.edition = edition
        self.bookshelf = get_remote_bookshelf(bookshelf)

    def long_version(self) -> str:
        """
        Long version identifier

        Of the form "{version}_e{edition}" e.g. v1.0.1_e002.

        Returns
        -------
        str
            Version identification string
        """
        return f"{self.version}_e{self.edition:03}"

    @staticmethod
    def path_parts(
        name: str,
        version: Version,
        edition: Edition,
        fname: Optional[str] = None,
    ) -> Iterable[str]:
        """
        Build the parts needed to unambiguously reference an edition.

        Parameters
        ----------
        name
            Book name
        version
            Book version
        edition
            Book edition
        fname
            If provided, reference a specific file within the data
        Returns
        -------
            Iterable of the parts, which can be joined as needed to get
            a path.
        """
        parts = [name, f"{version}_e{edition:03}"]
        if fname is not None:
            parts.append(fname)
        return parts

    @classmethod
    def relative_path(
        cls,
        name: str,
        version: Version,
        edition: Edition,
        fname: Optional[str] = None,
    ) -> str:
        """
        Build the relative path of the edition

        Parameters
        ----------
        name
            Book name
        version
            Book version
        edition
            Book edition
        fname
            If provided, reference a specific file within the data
        Returns
        -------
            Relative path to the book or resource within the book
        """
        return os.path.join(
            *cls.path_parts(name=name, version=version, edition=edition, fname=fname)
        )

    def url(self, fname: Optional[str] = None) -> str:
        """
        Get the expected URL for the book

        This URL is generated locally using the provided remote bookshelf

        Parameters
        ----------
        fname : str
            If provided get the URL of a file within the Book

        Returns
        -------
        str
            URL
        """
        return build_url(
            self.bookshelf,
            *self.path_parts(self.name, self.version, self.edition, fname),
        )


class LocalBook(_Book):
    """
    A local instance of a Book

    A Book consists of a metadata file (``datapackage.json``) and one or more ``Resource`` files.
    For now, these ``Resource's`` are only csv files of timeseries in the IAMC format, but
    this could be extended in future to handle additional data-types. Resources are fetched from
    the remote bookshelf when first requested and are cached locally for subsequent use.

    The ``Book`` metadata follow the ``datapackage`` specification with some additional metadata
    specific to this project. That means that each ``Book`` also doubles as a ``datapackage``.
    Once released by the ``Book`` author, a ``Book`` becomes immutable. If ``Book`` authors
    wish to update the metadata or data contained within a ``Book`` they must upload a new
    version of the ``Book``.
    """

    def __init__(
        self,
        name: str,
        version: str,
        edition: int = 1,
        local_bookshelf: Union[str, pathlib.Path, None] = None,
    ):
        super().__init__(name, version, edition)

        if local_bookshelf is None:
            local_bookshelf = create_local_cache(local_bookshelf)
        self.local_bookshelf = pathlib.Path(local_bookshelf)
        self._metadata: Optional[datapackage.Package] = None

    def hash(self) -> str:
        """
        Get the hash for the metadata

        This effectively also hashes the data as the metadata contains the hashes of
        the local Resource files.

        Returns
        -------
        str
            sha256 sum that is unique for the Book
        """
        return str(pooch.file_hash(self.local_fname(DATAPACKAGE_FILENAME)))

    def local_fname(self, fname: str) -> str:
        """
        Get the name of a file in the package

        Parameters
        ----------
        fname : str
            Name of the file

        Returns
        -------
        str
            The filename for the file in the local bookshelf
        """
        return os.path.join(self.local_bookshelf, self.name, self.long_version(), fname)

    def as_datapackage(self) -> datapackage.Package:
        """
        Package representation of the book

        :mod:`datapackage` is used for handling the metadata. Modifying
        the package also modifies the Book.

        Returns
        -------
        :class:`datapackage.Package`
            Metadata about the Book
        """
        if self._metadata is None:
            fname = DATAPACKAGE_FILENAME

            local_fname = self.local_fname(fname)
            with open(local_fname) as file_handle:
                file_data = json.load(file_handle)

            self._metadata = datapackage.Package(file_data)
        return self._metadata

    def metadata(self) -> Dict[str, Any]:
        """
        Metadata about the current book

        Returns
        -------
        dict
            Metadata about the Book
        """
        return cast(Dict[str, Any], self.as_datapackage().descriptor)

    def files(self) -> List[str]:
        """
        List of files that are locally available

        Since each Resource is fetched when first read the number of files present may
        be less than available on the remote bookshelf.

        Returns
        -------
        list of str
            List of paths of all Book's files, including `datapackage.json` which contains
            the metadata about the Book.
        """
        file_list = glob.glob(self.local_fname("*"))
        return file_list

    def add_timeseries(self, name: str, data: scmdata.ScmRun) -> None:
        """
        Add a timeseries resource to the Book

        Updates the Books metadata

        Parameters
        ----------
        name : str
            Unique name of the resource
        data : scmdata.ScmRun
            Timeseries data to add to the Book
        """
        fname = f"{name}.csv"
        data.timeseries().sort_index().to_csv(self.local_fname(fname))
        resource_hash = pooch.hashes.file_hash(self.local_fname(fname))

        metadata = self.as_datapackage()
        metadata.add_resource(
            {
                "name": name,
                "format": "CSV",
                "filename": fname,
                "hash": resource_hash,
            }
        )
        metadata.save(self.local_fname(DATAPACKAGE_FILENAME))

    @classmethod
    def create_new(
        cls, name: str, version: Version, edition: Edition = 1, **kwargs: Any
    ) -> "LocalBook":
        """
        Create a new Book
        """
        book = LocalBook(name, version, edition, **kwargs)
        book._metadata = datapackage.Package(
            {"name": name, "version": version, "edition": edition, "resources": []}
        )
        book._metadata.save(book.local_fname(DATAPACKAGE_FILENAME))

        return book

    @classmethod
    def create_from_metadata(cls, meta: NotebookMetadata, **kwargs: str) -> "LocalBook":
        """
        Create a new book from a notebook

        Parameters
        ----------
        meta : NotebookMetadata
            Metadata about the book

        kwargs
            Additional arguments passed to :class:`LocalBook`

        Returns
        -------
        LocalBook
            An instance of a local book with the datapackage setup
        """
        book = LocalBook(
            meta.name, version=meta.version, edition=meta.edition, **kwargs
        )
        book._metadata = datapackage.Package(
            {
                "name": meta.name,
                "version": meta.version,
                "private": meta.private,
                "edition": meta.edition,
                "resources": [],
            }
        )
        book._metadata.save(book.local_fname(DATAPACKAGE_FILENAME))

        return book

    def timeseries(self, name: str) -> scmdata.ScmRun:
        """
        Get a timeseries resource

        If the data is not available in the local cache, it is downloaded from the
        remote BookShelf.

        Parameters
        ----------
        name : str
            Name of the volume

        Returns
        -------
            Timeseries data

        """
        resource: datapackage.Resource = self.as_datapackage().get_resource(name)

        if resource is None:
            raise ValueError(f"Unknown timeseries '{name}'")

        local_fname = self.local_fname(resource.descriptor["filename"])
        fetch_file(
            self.url(resource.descriptor.get("filename")),
            pathlib.Path(local_fname),
            known_hash=resource.descriptor.get("hash"),
        )

        return scmdata.ScmRun(local_fname)



"""
publish CLI command
"""
import logging
import tempfile

import click

from bookshelf.notebook import get_available_versions, run_notebook
from bookshelf.shelf import BookShelf

logger = logging.getLogger(__name__)


@click.command("publish", short_help="Upload a book to the bookshelf")
@click.argument("name", required=True)
@click.option(
    "--version",
    multiple=True,
    help="List of versions to run",
    required=False,
)
@click.option(
    "--include-private/--no-include-private",
    help="Run private versions. These will likely fail if the data is not available locally",
    default=False,
)
@click.option(
    "--force",
    is_flag=True,
    help="Override the existing published data",
    default=False,
)
def cli(name: str, version: list[str], include_private: bool, force: bool) -> None:
    """
    Build and upload a Book to the Bookshelf

    Uploading a Book requires the correct AWS credentials (until an authentication
    scheme is introduced). At Climate Resource we use aws-vault for managing multiple
    sets of AWS credentials. Documentation about the different sources of authentication
    can be found here: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html

    To ensure a reproducible build, the Book is built from a notebook in an isolated
    output directory. There currently isn't any functionality to upload a pre-built Book.
    """
    if not version:
        all_versions = get_available_versions(name, include_private=include_private)
    else:
        all_versions = version

    for dataset_version in all_versions:
        with tempfile.TemporaryDirectory() as temp_dir:
            logger.info(f"Building Book in isolated environment: {temp_dir}")
            try:
                book = run_notebook(
                    name,
                    output_directory=temp_dir,
                    force=False,
                    version=dataset_version,
                )
                logger.info(f"Finish building {name}@{book.long_version()}")

                shelf = BookShelf()
                shelf.publish(book, force=force)
            except Exception as exc:
                logger.error(str(exc))
                raise click.Abort() from exc



import pytest

from bookshelf.constants import DEFAULT_BOOKSHELF
from bookshelf.utils import build_url, get_env_var, get_remote_bookshelf


@pytest.mark.parametrize(
    "root,paths,exp",
    (
        ("https://test.com", ["example"], "https://test.com/example"),
        ("https://test.com", ["example", "part2"], "https://test.com/example/part2"),
        ("https://test.com/sub", ["example"], "https://test.com/sub/example"),
        (
            "https://test:user@test.com",
            ["example"],
            "https://test:user@test.com/example",
        ),
    ),
)
def test_build_url(root, paths, exp):
    res = build_url(root, *paths)

    assert res == exp


def test_remote_bookshelf(monkeypatch):
    env_var = "BOOKSHELF_REMOTE"
    monkeypatch.delenv(env_var)
    assert get_remote_bookshelf(None) == DEFAULT_BOOKSHELF

    monkeypatch.setenv(env_var, "https://test.local")
    assert get_remote_bookshelf(None) == "https://test.local"

    assert get_remote_bookshelf("test") == "test"


def test_get_env_var(monkeypatch):
    exp_value = "true"
    monkeypatch.setenv("BOOKSHELF_TEST", exp_value)
    exp_err = "Environment variable BOOKSHELF_MISSING not set. Check configuration"

    with pytest.raises(ValueError, match=exp_err):
        get_env_var("missing")

    with pytest.raises(ValueError):
        get_env_var("test", add_prefix=False)

    assert get_env_var("test") == exp_value
    assert get_env_var("TeST") == exp_value
    assert get_env_var("BOOKSHELF_test", add_prefix=False) == exp_value



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Example
#
# This notebook is an example for the creation of a new book and can be used as a template.
#
# In this example, we fetch the RCMIP Radiative Forcing timeseries from `rcmip.org` and
# produce a resource that contains a filtered subset of variables.

# %%
import logging
import tempfile

import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# # Initialise

# %%
logging.basicConfig(level=logging.INFO)

# %%
metadata = load_nb_metadata("examples/simple")
metadata.dict()

# %% tags=["parameters"]
# This cell contains additional parameters that are controlled using papermill
local_bookshelf = tempfile.mkdtemp()

# %%
local_bookshelf

# %% [markdown]
# # Fetch

# %% [markdown]
# We recommend using `pooch` to download input data. `Pooch` will verify that the
# downloaded file matches the expected hash as a check that the download was performed
# successfully.

# %%
data_fname = metadata.download_file()
data = scmdata.ScmRun(data_fname, lowercase_cols=True)
data.head()

# %% [markdown]
# # Process

# %%
data.get_unique_meta("variable")

# %%
rf = data.filter(
    variable=["Radiative Forcing|Anthropogenic", "Radiative Forcing|Natural"]
)
rf

# %%
book = LocalBook.create_from_metadata(metadata, local_bookshelf=local_bookshelf)

# %% [markdown]
# Create a new `Resource` in the `Book` using the RF `scmdata.ScmRun` object. This function copies the timeseries to a local file and calculate the hash of this file. This hash can be used to check if the files have been modified.

# %%
book.add_timeseries("rf", rf)

# %% [markdown]
# Below the `Book`'s metadata is shown. This contains all of the metadata about the `Book` and the associated `Resources`.
#
# This is the metadata that clients download and can be used to fetch the `Book`'s `Resources`. Once deployed this `Book` becomes immutable. Any changes to the metadata or data requires releasing a new version of a `Book`.

# %%
book.metadata()

# %% [markdown]
# That is all.
#
# This notebook is not responsible for uploading the book to the `BookShelf`. See docs for how to upload `Books` to the `BookShelf`

# %%



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # ssp-basic-elements
# The basic elements from the SSP database. This includes country-level projections of population and GDP


# %%
import logging
import tempfile

import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# #  Initialise

# %%
logging.basicConfig(level="INFO")

# %%
metadata = load_nb_metadata("ssp-basic-elements")
metadata.dict()

# %% tags=["parameters"]
local_bookshelf = tempfile.mkdtemp()

# %%
local_bookshelf

# %%
book = LocalBook.create_from_metadata(metadata, local_bookshelf=local_bookshelf)

# %% [markdown]
# #  Fetch

# %%
metadata.download_file()
ssp_country_fname = metadata.download_file(0)
ssp_regions_fname = metadata.download_file(1)

# %%
basic_elements_country = scmdata.ScmRun(ssp_country_fname, lowercase_cols=True)
basic_elements_country

# %%
basic_elements_regions = scmdata.ScmRun(ssp_regions_fname, lowercase_cols=True)
basic_elements_regions

# %%
basic_elements_regions.get_unique_meta("variable")

# %% [markdown]
# # Process
#
# We don't need the population breakdowns so we filter them away

# %%
basic_elements_country = basic_elements_country.filter(
    variable=["Population", "Population|Urban|Share", "GDP|PPP"]
)
basic_elements_regions = basic_elements_regions.filter(
    variable=["Population", "Population|Urban|Share", "GDP|PPP"]
)

# %%
len(basic_elements_country.get_unique_meta("region"))


# %%
def clean_scenario(s):
    return s[:4]


basic_elements_country["scenario"] = basic_elements_country["scenario"].apply(
    clean_scenario
)
basic_elements_regions["scenario"] = basic_elements_regions["scenario"].apply(
    clean_scenario
)

# %%
basic_elements_regions.meta[["scenario", "model"]].drop_duplicates()

# %%
book.add_timeseries("by_country", basic_elements_country)
book.add_timeseries("by_region", basic_elements_regions)

# %%
book.metadata()

# %%



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

doc_source_dir = os.path.abspath(os.path.dirname(__file__))
src_dir = os.path.abspath(os.path.join(doc_source_dir, "..", "..", "src"))

sys.path.insert(0, src_dir)

project = "bookshelf"
copyright = "2023, Climate Resource"
author = "Jared Lewis"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",  # pass numpy style docstrings
    "sphinx_click.ext",
]

templates_path = ["_templates"]
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = "en"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = [os.path.join(doc_source_dir, "_static")]


# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "python": ("https://docs.python.org/3", None),
    "pyam": ("https://pyam-iamc.readthedocs.io/en/latest", None),
    "scmdata": ("https://scmdata.readthedocs.io/en/latest", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
    "pint": (
        "https://pint.readthedocs.io/en/latest",
        None,
    ),
}
napoleon_google_docstring = False
napoleon_numpy_docstring = True


# -- Misc configuration -------------------------------------------------
rst_epilog = r"""
.. |CO2| replace:: CO\ :sub:`2`\

.. |CH4| replace:: CH\ :sub:`5`\

.. |N2O| replace:: N\ :sub:`2`\O\
"""



"""
Bookshelf CLI
"""
import logging
import os

import click
import click_log
import dotenv

dotenv.load_dotenv()

cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))
logger = logging.getLogger("bookshelf")


class _CLICommands(click.MultiCommand):
    def list_commands(self, ctx):
        commands = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                commands.append(filename[4:-3])
        commands.sort()
        return commands

    def get_command(self, ctx, cmd_name):
        try:
            mod = __import__(f"bookshelf.commands.cmd_{cmd_name}", None, None, ["cli"])
        except ImportError:  # pragma: no cover
            return None
        return mod.cli


@click.command(cls=_CLICommands, name="bookshelf")
@click.option("-q", "--quiet", is_flag=True)
@click_log.simple_verbosity_option(logger)
@click.pass_context
def main(ctx, quiet):
    """
    Bookshelf for managing reusable datasets
    """
    ctx.ensure_object(dict)

    if not logger.hasHandlers():
        click_log.basic_config(logger)  # pragma: no cover

    logger.setLevel(logging.INFO)
    if quiet:
        logger.setLevel(logging.ERROR)


if __name__ == "__main__":
    main()



"""
Constants
"""
import os

DATA_FORMAT_VERSION = "v0.2.1"
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))

DATA_DIR = os.path.join(ROOT_DIR, "data")
RAW_DATA_DIR = os.path.join(DATA_DIR, "raw")
TEST_DATA_DIR = os.path.join(ROOT_DIR, "tests", "test-data")
PROCESSED_DATA_DIR = os.path.join(DATA_DIR, "processed", DATA_FORMAT_VERSION)


DEFAULT_BOOKSHELF = (
    f"https://cr-prod-datasets-bookshelf.s3.us-west-2.amazonaws.com"
    f"/{DATA_FORMAT_VERSION}"
)
ENV_PREFIX = "BOOKSHELF_"



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
import logging
import tempfile

import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# #  Initialise

# %%
logging.basicConfig(level="INFO")

# %%
metadata = load_nb_metadata("rcmip-emissions")
metadata.dict()

# %% tags=["parameters"]
local_bookshelf = tempfile.mkdtemp()

# %%
local_bookshelf

# %%
book = LocalBook.create_from_metadata(
    metadata,
    local_bookshelf=local_bookshelf,
)

# %% [markdown]
# #  Fetch

# %%
rcmip_fname = metadata.download_file()
rcmip_emissions = scmdata.ScmRun(rcmip_fname, lowercase_cols=True)
rcmip_emissions

# %% [markdown]
# # Process

# %%
rcmip_emissions.get_unique_meta("variable")

# %%
book.add_timeseries("complete", rcmip_emissions)


# %%
magicc_emissions = rcmip_emissions.filter(
    variable=[
        "Emissions|BC",
        "Emissions|CH4",
        "Emissions|CO",
        "Emissions|CO2",
        "Emissions|CO2|MAGICC AFOLU",
        "Emissions|CO2|MAGICC Fossil and Industrial",
        "Emissions|F-Gases|*",
        "Emissions|Montreal Gases|*",
        "Emissions|N2O",
        "Emissions|NH3",
        "Emissions|NOx",
        "Emissions|OC",
        "Emissions|Sulfur",
        "Emissions|VOC",
    ]
)

# %%
magicc_emissions.meta[["variable", "unit"]].drop_duplicates()

# %%
# for v in magicc_emissions.get_unique_meta("variable"):
#     plt.figure()
#     magicc_emissions.filter(variable=v, region="World").lineplot()

# %%
book.add_timeseries("magicc", magicc_emissions)

# %%
book.metadata()



"""
Bookshelf, Climate Resource's curated datasets
See README and docs for more info.
"""


try:
    from importlib.metadata import version as _version
except ImportError:  # pragma: no cover
    # no recourse if the fallback isn't there either...
    from importlib_metadata import version as _version  # type: ignore  # noqa

try:
    __version__ = _version("bookshelf")
except Exception:  # pragma: no cover pylint: disable=broad-except
    # Local copy, not installed with setuptools
    __version__ = "unknown"


from bookshelf.book import LocalBook  # noqa pylint: disable=wrong-import-position
from bookshelf.shelf import BookShelf  # noqa  pylint: disable=wrong-import-position



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Worldbank Data Indicators
#

# %%
import logging
import re
import tempfile
import zipfile

import pandas as pd
import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# # Initialise

# %%
logging.basicConfig(level=logging.INFO)

# %%
metadata = load_nb_metadata("wdi")
metadata.dict()

# %% tags=["parameters"]
# This cell contains additional parameters that are controlled using papermill
local_bookshelf = tempfile.mkdtemp()

# %%
local_bookshelf

# %% [markdown]
# # Fetch

# %% [markdown]
# We recommend using `pooch` to download input data. `Pooch` will verify that the
# downloaded file matches the expected hash as a check that the download was performed
# successfully.

# %%

# Data has been cached by CR (hash came from original download)
data_fname = metadata.download_file()

# %%
zf = zipfile.ZipFile(data_fname)
zf.filelist

# %%
df = pd.read_csv(zf.open("WDIData.csv"))

column_rename = {
    "Country Name": "name",
    "Country Code": "region",
    "Indicator Name": "variable",
    "Indicator Code": "variable_code",
}
df = df.rename(column_rename, axis=1)
df["scenario"] = "historical"
df["model"] = "World Bank"
df["source"] = f"WDI_{metadata.version}"
df["unit"] = ""
del df["Unnamed: 66"]

# %%
data = scmdata.ScmRun(df)
data.head()

# %%
unit_regex = re.compile(r"\s\((.*)\)")


def get_units(run):
    variable = run.get_unique_meta("variable", True)

    unit_match = re.search(unit_regex, variable)
    if unit_match:
        unit = unit_match.group(1)
        run["unit"] = unit
        variable = re.sub(unit_regex, "", variable)
    toks = variable.split(", ")
    variable = "|".join(
        [t.capitalize() if not t[0].isupper() else t for t in toks]
    ).rstrip("|")
    run["variable"] = variable

    return run


# This can take a minute or two
data = data.groupby("variable").map(get_units)

# %% [markdown]
# ## Emissions cleaning

# %%
data.filter(variable_code="EN.*").meta[["variable", "unit"]].drop_duplicates()

# %%
# Check units
data.meta[data.meta.unit.str.contains("kt")][["variable", "unit"]].drop_duplicates()

# %%
# Fix emissions units to be emissions/yr
data["unit"] = data["unit"].replace(
    "thousand metric tons of CO2 equivalent", "kt CO2-eq/yr"
)
data["unit"] = data["unit"].replace("kt of CO2 equivalent", "kt CO2-eq/yr")
data["unit"] = data["unit"].replace("Mt of CO2 equivalent", "Mt CO2-eq/yr")
# Check above shows that only emissions ts use "kt" as units
data["unit"] = data["unit"].replace("^kt$", "kt CO2/yr", regex=True)

# %%
# Rename variables
variable_map = {
    "Agricultural methane emissions": "Emissions|CH4|Agriculture",
    "Agricultural nitrous oxide emissions": "Emissions|N2O|Agriculture",
    "CO2 emissions": "Emissions|CO2",
    "CO2 emissions from electricity and heat production|Total": "Emissions|CO2",
    "CO2 emissions from gaseous fuel consumption": "Emissions|CO2|Gaseous Fuel Consumption",
    "CO2 emissions from liquid fuel consumption": "Emissions|CO2|Liquid Fuel Consumption",
    "CO2 emissions from manufacturing industries and construction": "Emissions|CO2|Manufacturing Industries and Construction",
    "CO2 emissions from other sectors|Excluding residential buildings and commercial and public services": "Emissions|CO2|Other Sectors",
    "CO2 emissions from residential buildings and commercial and public services": "Emissions|CO2|Residential Buildings and Commercial and Public Services",
    "CO2 emissions from solid fuel consumption": "Emissions|CO2|Solid Fuel Consumption",
    "CO2 emissions from transport": "Emissions|CO2|Transport",
    "Energy related methane emissions": "Emissions|CH4|Energy",
    "GHG net emissions/removals by LUCF": "Emissions|GHG|Net LUCF",
    "HFC gas emissions": "Emissions|HFCs",
    "Methane emissions": "Emissions|CH4",
    "Methane emissions in energy sector": "Emissions|CH4|Energy",
    "Nitrous oxide emissions": "Emissions|N2O",
    "Nitrous oxide emissions in energy sector": "Emissions|N2O|Energy",
    "Other greenhouse gas emissions": "Emissions|Other GHGs",
    "Other greenhouse gas emissions|HFC|PFC and SF6": "Emissions|HFCs|PFC and SF6",
    "PFC gas emissions": "Emissions|HFCs|PFC",
    "SF6 gas emissions": "Emissions|HFCs|SF6",
    "Total greenhouse gas emissions": "Emissions|GHG",
}

for old, new in variable_map.items():
    data["variable"] = data["variable"].replace(old, new)

# %%
data.filter(variable="Emissions*").meta[["variable", "unit"]].drop_duplicates()

# %% [markdown]
# ## Population

# %%
data.filter(variable="*Pop*").get_unique_meta("variable")

# %%
data.filter(variable="Population|*").meta[["variable", "unit"]].drop_duplicates()

# %%
data.get_unique_meta("variable")

# %%
data.filter(variable="GDP|PPP")

# %%
data.filter(variable="GDP|PPP", unit="constant 2017 international $").lineplot(
    units="region", estimator=None
)

# %% [markdown]
# # Process

# %%
data.get_unique_meta("region")

# %%
book = LocalBook.create_new(
    name=metadata.name, version=metadata.version, local_bookshelf=local_bookshelf
)

# %%
# Entire dataset (~168 MB uncompressed)
book.add_timeseries("clean", data)

# %%
data.filter(variable="GDP").get_unique_meta("unit")

# %%
# Smaller subset of data that is typically used for analysis
subset = scmdata.run_append(
    [
        data.filter(variable="GDP|PPP", unit="constant 2017 international $"),
        data.filter(variable="GDP", unit="constant 2015 US$"),
        data.filter(variable="Emissions|*"),
        data.filter(variable="Population|Total"),
    ]
)
book.add_timeseries("core", subset)

# %% [markdown]
# Below the `Book`'s metadata is shown. This contains all of the metadata about the `Book` and the associated `Resources`.
#
# This is the metadata that clients download and can be used to fetch the `Book`'s `Resources`. Once deployed this `Book` becomes immutable. Any changes to the metadata or data requires releasing a new version of a `Book`.

# %%
book.metadata()

# %% [markdown]
# That is all.
#
# This notebook is not responsible for uploading the book to the `BookShelf`. See docs for how to upload `Books` to the `BookShelf`

# %%



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Worldbank Data Indicators
#

# %%
import logging
import tempfile

import numpy as np
import pandas as pd
import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# # Initialise

# %%
logging.basicConfig(level=logging.INFO)

# %%
metadata = load_nb_metadata("un-wpp")
metadata.dict()

# %% tags=["parameters"]
# This cell contains additional parameters that are controlled using papermill
local_bookshelf = tempfile.mkdtemp()
local_bookshelf

# %%
book = LocalBook.create_new(
    metadata.name, version=metadata.version, local_bookshelf=local_bookshelf
)

# %% [markdown]
# # Fetch

# %% [markdown]
# We recommend using `pooch` to download input data. `Pooch` will verify that the
# downloaded file matches the expected hash as a check that the download was performed
# successfully.

# %%
data_fname = metadata.download_file()

# %%
df_hist = pd.read_excel(data_fname, sheet_name="Estimates", skiprows=16)
df_proj = pd.read_excel(data_fname, sheet_name="Medium variant", skiprows=16)

# %%
value_cols = [
    "Total Population, as of 1 January (thousands)",
    "Total Population, as of 1 July (thousands)",
    "Male Population, as of 1 July (thousands)",
    "Female Population, as of 1 July (thousands)",
    "Population Density, as of 1 July (persons per square km)",
    "Population Sex Ratio, as of 1 July (males per 100 females)",
    "Median Age, as of 1 July (years)",
    "Natural Change, Births minus Deaths (thousands)",
    "Rate of Natural Change (per 1,000 population)",
    "Population Change (thousands)",
    "Population Growth Rate (percentage)",
    "Population Annual Doubling Time (years)",
    "Births (thousands)",
    "Births by women aged 15 to 19 (thousands)",
    "Crude Birth Rate (births per 1,000 population)",
    "Total Fertility Rate (live births per woman)",
    "Net Reproduction Rate (surviving daughters per woman)",
    "Mean Age Childbearing (years)",
    "Sex Ratio at Birth (males per 100 female births)",
    "Total Deaths (thousands)",
    "Male Deaths (thousands)",
    "Female Deaths (thousands)",
    "Crude Death Rate (deaths per 1,000 population)",
    "Life Expectancy at Birth, both sexes (years)",
    "Male Life Expectancy at Birth (years)",
    "Female Life Expectancy at Birth (years)",
    "Life Expectancy at Age 15, both sexes (years)",
    "Male Life Expectancy at Age 15 (years)",
    "Female Life Expectancy at Age 15 (years)",
    "Life Expectancy at Age 65, both sexes (years)",
    "Male Life Expectancy at Age 65 (years)",
    "Female Life Expectancy at Age 65 (years)",
    "Life Expectancy at Age 80, both sexes (years)",
    "Male Life Expectancy at Age 80 (years)",
    "Female Life Expectancy at Age 80 (years)",
    "Infant Deaths, under age 1 (thousands)",
    "Infant Mortality Rate (infant deaths per 1,000 live births)",
    "Live Births Surviving to Age 1 (thousands)",
    "Under-Five Deaths, under age 5 (thousands)",
    "Under-Five Mortality (deaths under age 5 per 1,000 live births)",
    "Mortality before Age 40, both sexes (deaths under age 40 per 1,000 live births)",
    "Male Mortality before Age 40 (deaths under age 40 per 1,000 male live births)",
    "Female Mortality before Age 40 (deaths under age 40 per 1,000 female live births)",
    "Mortality before Age 60, both sexes (deaths under age 60 per 1,000 live births)",
    "Male Mortality before Age 60 (deaths under age 60 per 1,000 male live births)",
    "Female Mortality before Age 60 (deaths under age 60 per 1,000 female live births)",
    "Mortality between Age 15 and 50, both sexes (deaths under age 50 per 1,000 alive at age 15)",
    "Male Mortality between Age 15 and 50 (deaths under age 50 per 1,000 males alive at age 15)",
    "Female Mortality between Age 15 and 50 (deaths under age 50 per 1,000 females alive at age 15)",
    "Mortality between Age 15 and 60, both sexes (deaths under age 60 per 1,000 alive at age 15)",
    "Male Mortality between Age 15 and 60 (deaths under age 60 per 1,000 males alive at age 15)",
    "Female Mortality between Age 15 and 60 (deaths under age 60 per 1,000 females alive at age 15)",
    "Net Number of Migrants (thousands)",
    "Net Migration Rate (per 1,000 population)",
]


# %%
def prep_df(df, **kwargs):
    df = df[~df["Year"].isna()]

    df_wide = (
        df.melt(
            id_vars=[
                "Region, subregion, country or area *",
                "ISO3 Alpha-code",
                "Type",
                "Year",
            ],
            value_vars=value_cols,
        )
        .set_index(
            [
                "Region, subregion, country or area *",
                "ISO3 Alpha-code",
                "Type",
                "Year",
                "variable",
            ]
        )
        .unstack("Year")
    )
    na_values = "..."
    df_wide[df_wide == na_values] = np.nan

    df_wide.columns = df_wide.columns.get_level_values("Year").astype(int)
    df_wide = df_wide.reset_index().rename(
        {"Region, subregion, country or area *": "region", "Type": "type"}, axis=1
    )

    # Strip units from between brackets
    df_wide["unit"] = df_wide["variable"].str.extract(r"\((.*)\)")
    df_wide["variable"] = df_wide["variable"].str.extract(r"(.*) \(.*")

    df_wide["region"] = df_wide["region"].str.replace("WORLD", "World")

    for k in kwargs:
        df_wide[k] = kwargs[k]

    return df_wide.reset_index(drop=True)


prepped = prep_df(
    df_hist, model="World Population Prospects 2022", scenario="Historical"
)

prepped

# %%
data_hist = scmdata.ScmRun(
    prep_df(df_hist, model="World Population Prospects 2022", scenario="Historical")
)
data_proj = scmdata.ScmRun(
    prep_df(df_proj, model="World Population Prospects 2022", scenario="Medium variant")
)


# %% [markdown]
# ## Population
#
# Now we can slice the data set up into different views


# %%
data_hist

# %%
# Merge the historical data with the projections to create a single timeseries
data = scmdata.run_append([data_hist.set_meta("scenario", "Medium variant"), data_proj])
data["historical_end_year"] = data_hist["year"].iloc[-1]
data.filter(year=range(2019, 2025)).timeseries()

# %%
available_types = data.get_unique_meta("type")
available_types

# %%
for t in available_types:
    data_subset = data.filter(type=t)

    # Use iso code as region identifer for compatibility with other data
    if t == "Country/Area":
        data_subset["country"] = data_subset["region"]
        data_subset["region"] = data_subset["ISO3 Alpha-code"]

    book.add_timeseries(
        f'by_{t.replace(" ", "_").replace("/", "_").lower()}',
        data_subset.drop_meta("ISO3 Alpha-code"),
    )

# %%
book.metadata()

# %% [markdown]
# That is all.
#
# This notebook is not responsible for uploading the book to the `BookShelf`. See docs for how to upload `Books` to the `BookShelf`

# %%



"""
A BookShelf is a collection of Books that can be queried and fetched as needed.
"""
import json
import logging
import os
import pathlib
from typing import Iterable, List, Optional, Tuple, Union, cast

import boto3
import boto3.exceptions
import datapackage
import requests.exceptions

from bookshelf.book import LocalBook
from bookshelf.errors import UnknownBook, UnknownEdition, UnknownVersion, UploadError
from bookshelf.schema import BookVersion, Edition, Version, VolumeMeta
from bookshelf.utils import (
    build_url,
    create_local_cache,
    fetch_file,
    get_env_var,
    get_remote_bookshelf,
)

logger = logging.getLogger(__name__)


def _fetch_volume_meta(
    name: str,
    remote_bookshelf: str,
    local_bookshelf: pathlib.Path,
    force: bool = True,
) -> VolumeMeta:
    """
    Fetch information about the books available for a given volume

    Parameters
    ----------
    name : str
        Name of the volume to fetch
    remote_bookshelf : str
        URL for the remote bookshelf
    local_bookshelf : pathlib.Path
        Local path where downloaded books will be stored.

        Must be a writable directory
    force: bool
        If True metadata is always fetched from the remote bookshelf

    Returns
    -------
    VolumeMeta
    """
    fname = "volume.json"

    local_fname = local_bookshelf / name / fname
    url = build_url(remote_bookshelf, name, fname)

    fetch_file(url, local_fname, force=force)

    with open(str(local_fname)) as file_handle:
        data = json.load(file_handle)

    return VolumeMeta(**data)


def _upload_file(s3, bucket, key, fname):  # pylint: disable=invalid-name
    try:
        logger.info(f"Uploading {fname} to {bucket} - {key}")
        s3.upload_file(fname, bucket, key, ExtraArgs={"ACL": "public-read"})
    except boto3.exceptions.S3UploadFailedError as s3_error:
        logger.exception(s3_error, exc_info=False)
        raise UploadError(f"Failed to upload {fname} to s3") from s3_error


def _update_volume_meta(book: LocalBook, remote_bookshelf: str) -> str:
    try:
        volume_meta = _fetch_volume_meta(
            book.name, remote_bookshelf, book.local_bookshelf, force=True
        )
    except requests.exceptions.HTTPError:
        volume_meta = VolumeMeta(name=book.name, license="", versions=[])

    meta_fname = str(book.local_bookshelf / book.name / "volume.json")
    volume_meta.versions.append(
        BookVersion(
            **{
                "version": book.version,
                "edition": book.edition,
                "url": book.url(),
                "hash": book.hash(),
                "private": book.metadata().get("private", False),
            }
        )
    )
    with open(meta_fname, "w") as file_handle:
        file_handle.write(volume_meta.json())

    return meta_fname


class BookShelf:
    """
    A BookShelf stores a number of Books

    If a Book isn't available locally, it will be queried from the remote bookshelf.

    Books can be fetched using :func:`load` by name. Specific versions of a book can be
    pinned if needed otherwise the latest version of the book is loaded.
    """

    def __init__(
        self,
        path: Union[str, pathlib.Path, None] = None,
        remote_bookshelf: Optional[str] = None,
    ):
        if path is None:
            path = create_local_cache(path)
        self.path = pathlib.Path(path)
        self.remote_bookshelf = get_remote_bookshelf(remote_bookshelf)

    def load(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
        force: bool = False,
    ) -> LocalBook:
        """
        Load a book

        If the book's metadata does not exist locally or an unknown version is requested
        the remote bookshelf is queried, otherwise the local metadata is used.

        Parameters
        ----------
        name: str
            Name of the volume to load
        version: str
            Version to load

            If no version is provided, the latest version is returned
        edition: int
            Edition of book to load

            If no edition is provided, the latest edition of the selected version is returned
        force: bool
            If True, redownload the book metadata

        Raises
        ------
        UnknownVersion
            The requested version is not available for the selected volume
        UnknownBook
            An invalid volume is requested

        Returns
        -------
        :class:`LocalBook`
            A book from which the resources can be accessed
        """
        if version is None or edition is None or force:
            version, edition = self._resolve_version(name, version, edition)

        metadata_fragment = LocalBook.relative_path(
            name, version, edition, "datapackage.json"
        )
        metadata_fname = self.path / metadata_fragment

        if not metadata_fname.exists():
            try:
                url = build_url(
                    self.remote_bookshelf,
                    *LocalBook.path_parts(name, version, edition, "datapackage.json"),
                )
                fetch_file(
                    url,
                    local_fname=metadata_fname,
                    known_hash=None,
                    force=force,
                )
            except requests.exceptions.HTTPError as http_error:
                raise UnknownVersion(name, version) from http_error

        if not metadata_fname.exists():
            raise AssertionError()  # noqa

        return LocalBook(name, version, edition, local_bookshelf=self.path)

    def is_available(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
    ) -> bool:
        """
        Check if a Book is available from the remote bookshelf

        Parameters
        ----------
        name : str
            Name of the volume to check
        version : str
            Version of the volume to check

            If no version is provided, then check if any Book's with a matching name
            have been uploaded.

        Returns
        -------
        bool
            True if a Book with a matching name and version exists on the remote bookshelf
        """
        try:
            self._resolve_version(name, version, edition)
            return True
        except (UnknownBook, UnknownVersion, UnknownEdition):
            return False

    def is_cached(self, name: str, version: Version, edition: Edition) -> bool:
        """
        Check if a book with a matching name/version is cached on the local bookshelf

        Parameters
        ----------
        name : str
            Name of the volume to check
        version : str
            Version of the volume to check
        edition : int
            Edition of the volume to check

        Returns
        -------
        bool
            True if a matching book is cached locally
        """
        try:
            # Check if the metadata for the book can be successfully read
            book = LocalBook(name, version, edition, local_bookshelf=self.path)
            book.metadata()
            return True
        except FileNotFoundError:
            return False

    def publish(self, book: LocalBook, force: bool = False) -> None:
        """
        Publish a book to the remote bookshelf

        Parameters
        ----------
        book : LocalBook
            Book to upload
        force : bool
            If True, overwrite any existing data

        Raises
        ------
        UploadError
            Unable to upload the book to the remote bookshelf. See error message for
            more information about how to resolve this issue.
        """
        if self.is_available(book.name, book.version):
            remote_book = self.load(book.name, book.version)

            if remote_book.edition >= book.edition:
                msg = (
                    f"Edition value has not been increased "
                    f"(remote: {remote_book.long_version()}, local: {book.long_version()})"
                )
                if not force:
                    raise UploadError(msg)
                logger.error(msg)
            logger.warning("Uploading a new edition of an existing book")
        files = book.files()

        # Check if additional files are going to be uploaded
        resources = book.as_datapackage().resources
        resource_fnames = [
            resource.descriptor["filename"]
            for resource in cast(Iterable[datapackage.Resource], resources)
        ]
        for resource_file in files:
            fname = os.path.basename(resource_file)
            if fname == "datapackage.json":
                continue
            if fname not in resource_fnames:
                raise UploadError(f"Non-resource file {fname} found in book")

        # Upload using boto3 by default for testing
        # Maybe support other upload methods in future

        s3 = boto3.client("s3")  # pylint: disable=invalid-name
        bucket = get_env_var("BUCKET", add_prefix=True)
        prefix = get_env_var("BUCKET_PREFIX", add_prefix=True)

        logger.info(f"Beginning to upload {book.name}@{book.version}")
        for resource_file in files:
            key = "/".join(
                (
                    prefix,
                    book.name,
                    book.long_version(),
                    os.path.basename(resource_file),
                )
            )
            _upload_file(s3, bucket, key, resource_file)

        # Update the metadata with the latest version information
        # Note that this doesn't have any guardrails and is susceptible to race conditions
        # Shouldn't be a problem for testing, but shouldn't be used in production
        meta_fname = _update_volume_meta(book, self.remote_bookshelf)
        key = "/".join((prefix, book.name, os.path.basename(meta_fname)))
        _upload_file(s3, bucket, key, meta_fname)

        logger.info(
            f"Book {book.name}@{book.version} ed.{book.edition} uploaded successfully"
        )

    def _resolve_version(
        self,
        name: str,
        version: Optional[Version] = None,
        edition: Optional[Edition] = None,
    ) -> Tuple[Version, Edition]:
        # Update the package metadata
        try:
            meta = _fetch_volume_meta(name, self.remote_bookshelf, self.path)
        except requests.exceptions.HTTPError as http_error:
            raise UnknownBook(f"No metadata for {repr(name)}") from http_error

        if version is None:
            version = meta.get_latest_version()

        # Verify that the version exists
        matching_version_books = meta.get_version(version)
        if not matching_version_books:
            raise UnknownVersion(name, version)

        # Find edition
        if edition is None:
            edition = matching_version_books[-1].edition
        if edition not in [b.edition for b in matching_version_books]:
            raise UnknownEdition(name, version, edition)
        return version, edition

    def list_versions(self, name: str) -> List[str]:
        """
        Get a list of available versions for a given Book

        Parameters
        ----------
        name: str
            Name of book

        Returns
        -------
        list of str
            List of available versions
        """
        try:
            meta = _fetch_volume_meta(name, self.remote_bookshelf, self.path)
        except requests.exceptions.HTTPError as http_error:
            raise UnknownBook(f"No metadata for {repr(name)}") from http_error

        return [version.version for version in meta.versions if not version.private]

    def list_books(self) -> List[str]:
        """
        Get a list of book names

        Returns
        -------
        list of str
            List of available books
        """
        raise NotImplementedError



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # PRIMAP-hist
#

# %%
# Need to think about how to handle notebook dependencies
# !pip install pycountry

# %%
import logging
import re
import tempfile

import pandas as pd
import pycountry
import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# # Initialise

# %%
logging.basicConfig(level=logging.INFO)

# %% tags=["parameters"]
# This cell contains additional parameters that are controlled using papermill
local_bookshelf = tempfile.mkdtemp()
version = "v2.4"

# %%
metadata = load_nb_metadata("primap-hist", version=version)
metadata.dict()


# %% [markdown]
# # Fetch

# %% [markdown]
# We recommend using `pooch` to download input data. `Pooch` will verify that the
# downloaded file matches the expected hash as a check that the download was performed
# successfully.

# %%
data_df = pd.read_csv(metadata.download_file())
data_df

# %%
col_renames = {
    "scenario (PRIMAP-hist)": "scenario",
    "entity": "variable",
    "area (ISO3)": "region",
    "country": "region",  # v2.2
    "category (IPCC2006_PRIMAP)": "category",
}
data_df_renamed = data_df.rename(col_renames, axis=1)
data_df_renamed["model"] = "PRIMAP-hist"
data_df_renamed["variable"] = "Emissions|" + data_df_renamed["variable"]

# %%
data = scmdata.ScmRun(data_df_renamed, lowercase_cols=True)
data.head()


# %% [markdown]
# # Process
#
# Minor renames and enrichment


# %%
def rename_variable(v):
    return (
        v.replace("KYOTOGHG", "Kyoto GHG")
        .replace("HFCS", "HFCs")
        .replace("PFCS", "PFCs")
        .replace("FGASES", "F-Gases")
    )


data["variable"] = data["variable"].apply(rename_variable)

# %%
data["scenario"] = data["scenario"].str.replace("HISTCR", "Historical|Country Reported")
data["scenario"] = data["scenario"].str.replace("HISTTP", "Historical|Third Party")

# %%
# Extract the GWP context from brackets
pattern = re.compile(r".*\((.*)\)")
pattern_all_but = re.compile(r"(.*) \(.*")

m = re.match(pattern_all_but, "Emissions|Kyoto GHG (AR4GWP100)")


def add_gwp_context(v):
    m = re.match(pattern, v)
    if m:
        return m.group(1)
    return None


def remove_gwp_from_variable(v):
    m = re.match(pattern_all_but, v)
    if m:
        return m.group(1)
    return v


# %%
data["gwp_context"] = data["variable"].apply(add_gwp_context)
data["variable"] = data["variable"].apply(remove_gwp_from_variable)

# %%
data.get_unique_meta("gwp_context")

# %%
data.get_unique_meta("variable")

# %%
data.timeseries()

# %%
region_map = {
    "ANNEXI": "Annex I",
    "ANT": "Antartica",
    "AOSIS": "Alliance of Small Island States (AOSIS)",
    "BASIC": "BASIC",
    "EARTH": "World",
    "EU27BX": "EU27 Post-Brexit",
    "LDC": "Least Developed Countries (LDC)",
    "NONANNEXI": "Non-Annex I",
    "UMBRELLA": "Umbrella",
}


# %%
def rename_regions(d):
    region = d.get_unique_meta("region", True)

    if region in region_map:
        region = region_map[region]
        d["region"] = region

    country_data = pycountry.countries.get(alpha_3=region)
    if country_data is not None:
        d["country"] = country_data.name

    return d


# Rename regions
data = data.groupby("region").map(rename_regions)


regions = []
for region in data.get_unique_meta("region"):
    country_data = pycountry.countries.get(alpha_3=region)
    if country_data is None:
        regions.append(region)
        print(region)

# %%
data.get_unique_meta("unit")


# %%
def convert_units(run):
    unit = run.get_unique_meta("unit", True)

    return run.convert_unit(unit.replace("Gg", "kt").replace("gigagram", "kt"))


data = data.groupby("unit").map(convert_units)

# %%
data.get_unique_meta("unit")

# %%
data.get_unique_meta("category")

# %%
data.timeseries()

# %%
data_countries = data.filter(region=regions, keep=False)
data_regions = data.filter(region=regions).drop_meta("country")
data_regions

# %%
data_countries.get_unique_meta("region")

# %%
book = LocalBook.create_from_metadata(metadata, local_bookshelf=local_bookshelf)

# %%
book.add_timeseries("by_country", data_countries)
book.add_timeseries("by_regions", data_regions)

# %%
book.metadata()

# %% [markdown]
# That is all.
#
# This notebook is not responsible for uploading the book to the `BookShelf`. See docs for how to upload `Books` to the `BookShelf`

# %%
